C===============================================================
C Parallel MG solver for 3D Poisson equation with Dirichlet BC
C using PVM 3 library calls.
C
C The computational domain is (0,A)x(0,B)x(0,C)
C The finest grid is (N+1)x(M+1)x(L+1).
C
C This program uses Correction Storage scheme, red-black
C Gauss-Seidel relaxation, half weighting for fine-to-coarse
C residual transfer, and bilinear interpolation for
C coarse-to-fine prolongation.
C
C This implementation does not use group functions of PVM 3.
C===============================================================
C
      INCLUDE 'pvmp3d.cmm'
C
C Allocate U & F arrays.
      DIMENSION UF(LENUF)
C
      INTEGER SPAWN11
C
C Get my tid and my parent's tid
      CALL PVMFMYTID( MYTID )
      IF (MYTID.LT.0) THEN
         PRINT *,'Can not find my TID!'
         CALL PVMFEXIT(INFO)
         STOP
      ENDIF
      CALL PVMFPARENT( PARENTID )
C
      IF ( NREALSIZE .EQ. 4 ) THEN
         NREALTYPE=REAL4
      ELSE
         NREALTYPE=REAL8
      ENDIF
C
C If I am the first instance (parentid<0) then get parameters
C and spawn other tasks (TOL=tolerance on the  residual).
      IF ( PARENTID .LT. 0 )  THEN
         TIDS(0)=MYTID
         OPEN(10,FILE='data')
         READ(10,*) N,M,L
         READ(10,*) A,B,C
         READ(10,*) TOL
         READ(10,*) NPROCX,NPROCY,NPROCZ
         NPROC=NPROCX*NPROCY*NPROCZ
         IF (NPROCX.LE.0 .OR. NPROCY.LE.0 .OR. NPROCZ.LE.0 .OR.
     1       NPROC.GT.MAXPROC) THEN
            CLOSE(10)
            CALL EXIT(1)
         ENDIF
         READ(10,*) NROUTE
         READ(10,*) NENCODING
         CLOSE(10)
C
C Spawn child tasks and pass the parameters to them
         SPEEDS(0)=SPAWN11('pvmp3d ',NPROC-1,TIDS(1),SPEEDS(1),
     1                     INFO)
***         CALL PVMFSPAWN('pvmp3d ',PVMDEFAULT,'*',NPROC-1,
***     1                     TIDS(1),INFO)
         IF (INFO.NE.NPROC-1) CALL EXIT(2)
         IF (NPROC.GT.1) THEN
C
C Send parameters to all child tasks
            MSGTAG=1001
            CALL PVMFINITSEND(PVMDEFAULT,INFO)
            IF (INFO.LT.0) CALL EXIT(3)
            CALL PVMFPACK(INTEGER4,NPROC,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(4)
            CALL PVMFPACK(INTEGER4,N,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(8)
            CALL PVMFPACK(INTEGER4,M,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(9)
            CALL PVMFPACK(INTEGER4,L,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(10)
            CALL PVMFPACK(NREALTYPE,A,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(11)
            CALL PVMFPACK(NREALTYPE,B,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(12)
            CALL PVMFPACK(NREALTYPE,C,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(13)
            CALL PVMFPACK(NREALTYPE,TOL,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(14)
            CALL PVMFPACK(INTEGER4,NPROCX,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(15)
            CALL PVMFPACK(INTEGER4,NPROCY,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(16)
            CALL PVMFPACK(INTEGER4,NPROCZ,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(17)
            CALL PVMFPACK(INTEGER4,TIDS,NPROC,1,INFO)
            IF (INFO.LT.0) CALL EXIT(18)
            CALL PVMFPACK(INTEGER4,NROUTE,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(5)
            CALL PVMFPACK(INTEGER4,NENCODING,1,1,INFO)
            IF (INFO.LT.0) CALL EXIT(19)
            CALL PVMFMCAST(NPROC-1,TIDS(1),MSGTAG,INFO)
            IF (INFO.LT.0) CALL EXIT(20)
         ENDIF
         PRINT 1002,NPROC, NPROCX, NPROCY, NPROCZ,
     1              (TIDS(I), SPEEDS(I), I=0,NPROC-1)
         PRINT 1003,N,M,L,A,B,C,TOL
      ELSE
C
C Get parameters from the master task
         MSGTAG=1001
         CALL PVMFRECV(PARENTID,MSGTAG,INFO)
         IF (INFO.LT.0) CALL EXIT(21)
         CALL PVMFUNPACK(INTEGER4,NPROC,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(22)
         CALL PVMFUNPACK(INTEGER4,N,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(25)
         CALL PVMFUNPACK(INTEGER4,M,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(26)
         CALL PVMFUNPACK(INTEGER4,L,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(27)
         CALL PVMFUNPACK(NREALTYPE,A,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(28)
         CALL PVMFUNPACK(NREALTYPE,B,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(29)
         CALL PVMFUNPACK(NREALTYPE,C,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(30)
         CALL PVMFUNPACK(NREALTYPE,TOL,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(31)
         CALL PVMFUNPACK(INTEGER4,NPROCX,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(32)
         CALL PVMFUNPACK(INTEGER4,NPROCY,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(33)
         CALL PVMFUNPACK(INTEGER4,NPROCZ,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(34)
         CALL PVMFUNPACK(INTEGER4,TIDS,NPROC,1,INFO)
         IF (INFO.LT.0) CALL EXIT(35)
         CALL PVMFUNPACK(INTEGER4,NROUTE,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(23)
         CALL PVMFUNPACK(INTEGER4,NENCODING,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(36)
      ENDIF
C
C Find my entry in the TIDS table
      DO 10 I=0,NPROC-1
         IF (TIDS(I).EQ.MYTID) THEN
            ME=I
            GOTO 11
         ENDIF
10    CONTINUE
C Can't find my self in the TIDS array (it should never happen)
      CALL EXIT(37)
11    CONTINUE
C
C Setup routing option for PVM 3
C 1=PvmDontRoute, 2=PvmAllowDirect (default), 3=PvmRouteDirect
      CALL PVMFSETOPT(PVMROUTE, NROUTE, IOLD)
      IF (IOLD.LT.0) CALL EXIT(38)
C
C Init. MG parameters and check for the size of allocated array
      LEN1=INITMG(N,M,L)
      I=((LEN1+LEN1)*NREALSIZE+1023)/1024
      IF (NREALTYPE.EQ.REAL8) THEN
         PRINT 1004,'Double Precision',NROUTE,NENCODING,I
      ELSE IF (NREALTYPE.EQ.REAL4) THEN
         PRINT 1004,'Single Precision',NROUTE,NENCODING,I
      ENDIF
      IF (LEN1+LEN1.GT.LENUF) THEN
         PRINT *,'INSUFFICIENT ARRAY SIZE (INCREASE "LENUF")!'
         CALL EXIT(40)
      ENDIF
C
C Display the subgrids allocated to this task.
C (on all grid levels).
      DO 1 I=1,NLVLS
         PRINT 1001,N/2**(I-1),M/2**(I-1),L/2**(I-1),
     1   ISTART(I),IEND(I), JSTART(I),JEND(I), KSTART(I),KEND(I)
1     CONTINUE
C
C The setup stage is completed at this point
      IF (ME.EQ.0) THEN
         PRINT *,'=============='
         PRINT *,'Setup complete'
         PRINT *,'=============='
      ENDIF
C
C Do computations
      CALL SUB(UF(1),UF(LEN1+1),N,M,L,A,B,C,ISTART(1),IEND(1),
     1         JSTART(1),JEND(1),KSTART(1),KEND(1),TOL)
C
C Program finished. Leave group and exit PVM
      CALL EXIT(0)
C
1001  FORMAT(' ',I3,'   ',I3,'   ',I3,'   (',I3,',',I3,
     1       ')   (',I3,',',I3,')   (',I3,',',I3,')')
1002  FORMAT(' TIDS (',I2,' task(s),  NPROCX=',I2,
     1       '  NPROCY=',I2,'  NPROCZ=',I2,')',/,
     2       (8X,'TID=',Z8,'     SPEED=',I6))
1003  FORMAT(1X,'N=',I4,'   M=',I4,'   L=',I4,/,
     1       1X,'A=',1PG12.5,'   B=',1PG12.5,'   C=',1PG12.5,
     2       '   TOL=',1PG12.5)
1004  FORMAT(1X,A,', pvm_setopt(PVMROUTE,',I1,',IOLD), ',
     1       'Encoding=',I1,/,
     2       ' U+F size required by this task =',I6,'K bytes')
      END
C
C---------------------------------------------------------------
C
      SUBROUTINE EXIT(IERR)
C
C Largest error number used: 71 (modify this number!)
C We don't intend to intepret the error numbers but simply print
C them. It's easy to locate the error for a given error number.
      IF (IERR.NE.0) PRINT 1000,IERR
      CALL PVMFEXIT(INFO)
      STOP
1000  FORMAT(1X,'*** Error code =',I3)
      END
C
C---------------------------------------------------------------
C
      INTEGER FUNCTION INITMG(N,M,L)
C
C Init. parameters for the MG solver.
C Return maxi. array size required.
      INCLUDE 'pvmp3d.cmm'
C
C Compute MEX, MEY and MEZ s.t. ME=MEX+(MEY+MEZ*NPROCX)*NPROCY
      I=ME
      MEX=MOD(I,NPROCX)
      I=I/NPROCX
      MEY=MOD(I,NPROCY)
      I=I/NPROCY
      MEZ=I
C
C Compute # of possible grid levels, and the subgrids allocated
C to me on all grid levels.
      NLVLS=1
      N1=N
      M1=M
      L1=L
      LEN=0
1     CONTINUE
C
C The subgrid allocated to me is (IS,IE)x(JS,JE)x(KS,KE),
C and the U & F arrays are defined as:
C U(IS:IE,JS:JE,KS:KE), F(IS:IE,JS:JE,KS:KE)
C
C The subgrids allocated to each task are defined in such a way
C that:
C
C   1. the set of inner points may be empty, but its size is
C      at least one grid step size in each direction (2 nodes).
C
C   2. two neighbouring subgrids overlap exactly 1 grid step.
C
C In case of uniform distribution (i.e., all subgrids have
C approximately the same size), IS, etc., can be computed by:
C     IS+1 = smallest i such that i >  MEX*N1/NPROCX
C     IE-1 = largest  i such that i <= (MEX+1)*N1/NPROCX
C
C     JS+1 = smallest j such that j >  MEY*M1/NPROCY
C     JE-1 = largest  j such that j <= (MEY+1)*M1/NPROCY
C
C     KS+1 = smallest k such that k >  MEZ*L1/NPROCZ
C     KE-1 = largest  k such that k <= (MEZ+1)*L1/NPROCZ
C
      IS=(MEX*N1)/NPROCX+1-1
      IF (MEX.EQ.NPROCX-1) THEN
         IE=((MEX+1)*N1)/NPROCX
      ELSE
         IE=((MEX+1)*N1)/NPROCX+1
      ENDIF
C
      JS=(MEY*M1)/NPROCY+1-1
      IF (MEY.EQ.NPROCY-1) THEN
         JE=((MEY+1)*M1)/NPROCY
      ELSE
         JE=((MEY+1)*M1)/NPROCY+1
      ENDIF
C
      KS=(MEZ*L1)/NPROCZ+1-1
      IF (MEZ.EQ.NPROCZ-1) THEN
         KE=((MEZ+1)*L1)/NPROCZ
      ELSE
         KE=((MEZ+1)*L1)/NPROCZ+1
      ENDIF
C
      LEN=LEN+(IE-IS+1)*(JE-JS+1)*(KE-KS+1)
      ISTART(NLVLS)=IS
      IEND(NLVLS)=IE
      JSTART(NLVLS)=JS
      JEND(NLVLS)=JE
      KSTART(NLVLS)=KS
      KEND(NLVLS)=KE
C
C Check if further coarsenning is possible
      IF ( MOD(N1,2).EQ.0 .AND. MOD(M1,2).EQ.0 .AND.
     1     MOD(L1,2).EQ.0 .AND.
     2     N1.GT.2 .AND. M1.GT.2 .AND. L1.GT.2 .AND.
     3     NLVLS.LT.MAXLVL) THEN
         NLVLS=NLVLS+1
         N1=N1/2
         M1=M1/2
         L1=L1/2
         GOTO 1
      ENDIF
C
C Return total array size (for U or F) required
      INITMG=LEN
      RETURN
      END
C
C---------------------------------------------------------------
C
      SUBROUTINE SUB(U,F,N,M,L,A,B,C,IS,IE,JS,JE,KS,KE,TOL)
      INCLUDE 'pvmp3d.cmm'
      DIMENSION U(IS:IE,JS:JE,KS:KE),F(IS:IE,JS:JE,KS:KE)
      DOUBLE PRECISION TM0(3),TM1(3), DMFLU, DMFLS
C
C The analytic solution UXYZ(x,y,z) and rhs FXYZ(x,y,z).
C UXYZ is used to define the boundary values for U,
C and to check the accuracy of the numerical solution obtained.
      UXYZ(X,Y,Z)=10.0*(X*X*X+Y*Y*Y+Z*Z*Z)
      FXYZ(X,Y,Z)=-60.0*(X+Y+Z)
C
C Set initial solution to 0 and assign right-hand-side values
      DO 10 K=KS,KE
         Z=K*C/L
         DO 10 J=JS,JE
            Y=J*B/M
            DO 10 I=IS,IE
               X=I*A/N
               U(I,J,K)=0.0
               F(I,J,K)=FXYZ(X,Y,Z)
10    CONTINUE
C
C Assign boundary values of U
C ---- x=0 or x=A
      IF (IS.EQ.0.OR.IE.EQ.N) THEN
         DO 20 K=KS,KE
            Z=K*C/L
            DO 20 J=JS,JE
               Y=J*B/M
               IF (IS.EQ.0) U(0,J,K)=UXYZ(0.0,Y,Z)
               IF (IE.EQ.N) U(N,J,K)=UXYZ(A,Y,Z)
20       CONTINUE
      ENDIF
C ---- y=0 or y=B
      IF (JS.EQ.0.OR.JE.EQ.M) THEN
         DO 21 K=KS,KE
            Z=K*C/L
            DO 21 I=IS,IE
               X=I*A/N
               IF (JS.EQ.0) U(I,0,K)=UXYZ(X,0.0,Z)
               IF (JE.EQ.M) U(I,M,K)=UXYZ(X,B,Z)
21       CONTINUE
      ENDIF
C ---- z=0 or z=C
      IF (KS.EQ.0.OR.KE.EQ.L) THEN
         DO 22 J=JS,JE
            Y=J*B/M
            DO 22 I=IS,IE
               X=I*A/N
               IF (KS.EQ.0) U(I,J,0)=UXYZ(X,Y,0.0)
               IF (KE.EQ.L) U(I,J,L)=UXYZ(X,Y,C)
22       CONTINUE
      ENDIF
C
C Assign parameters for POIS3D
      NMIN=2
      MMIN=2
      LMIN=2
      MAXIT=10
      IREL=0
      CALL GETTIME(TM0)
      CALL TIMING(0,N,M,L)
      CALL POIS3D(U,F,N,M,L,A,B,C,NMIN,MMIN,LMIN,
     1            MAXIT,TOL,RESIDU,IREL)
      CALL TIMING(-1,N,M,L)
      CALL GETTIME(TM1)
C
C Compare numerical solution to the analytical one.
      ERROR=0.0
      DO 30 K=KS+1,KE-1
         Z=K*C/L
         DO 30 J=JS+1,JE-1
            Y=J*B/M
            DO 30 I=IS+1,IE-1
               X=I*A/N
               ERROR=MAX(ERROR,ABS(U(I,J,K)-UXYZ(X,Y,Z)))
30    CONTINUE
C
C Compute the maximum of errors obtained on all tasks
      IF (ME.EQ.0) THEN
C
C Gather approximation errors from all child-tasks.
         IF (NPROC.GT.1) THEN
            MSGTAG=6
            DO 50 I=1,NPROC-1
               CALL PVMFRECV(-1,MSGTAG,INFO)
               IF (INFO.LT.0) CALL EXIT(44)
               CALL PVMFUNPACK(NREALTYPE,E,1,1,INFO)
               IF (INFO.LT.0) CALL EXIT(45)
               ERROR=MAX(ERROR,E)
50          CONTINUE
         ENDIF
         PRINT *,'Error = ',ERROR
      ELSE
C
C Send my part of error to the master task
         CALL PVMFINITSEND(PVMDEFAULT,INFO)
         IF (INFO.LT.0) CALL EXIT(46)
         CALL PVMFPACK(NREALTYPE,ERROR,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(47)
         MSGTAG=6
         CALL PVMFSEND(PARENTID,MSGTAG,INFO)
         IF (INFO.LT.0) CALL EXIT(48)
      ENDIF
C
C Print out some statistics on CPU time, work amount, etc.
      CNTTMP=(CNTADD+CNTMUL)*1D-6
      IF (DABS(TM1(1)-TM0(1))+DABS(TM1(3)-TM0(3)).LT.1D-5) THEN
         DMFLU=0.0D0
         DMFLS=0.0D0
      ELSE
         DMFLU=CNTTMP/(TM1(1)-TM0(1))
         DMFLS=CNTTMP/(TM1(3)-TM0(3))
      ENDIF
      PRINT 1000, CNTADD*1D-6, CNTMUL*1D-6, CNTTMP, ME,
     1            TM1(1)-TM0(1), TM1(2)-TM0(2), TM1(3)-TM0(3),
     2            DMFLU, DMFLS
C
1000  FORMAT(' Number of million operations:',/,
     1              '   Add/Sub:',1PG12.5,
     2              '   Mul/Div:',1PG12.5,
     3              '   Total:  ',1PG12.5,/,
     4       ' Processing time for ME=',I3,' (seconds):',/,
     5              '   User:   ',1PG12.5,
     6              '   System: ',1PG12.5,
     7              '   Real:   ',1PG12.5,/,
     8       ' Performance (MFLOPS):',/,
     9              '   /User:  ',1PG12.5,
     A              '   /Real:  ',1PG12.5)
      RETURN
      END
C
C...............................................................
C
      SUBROUTINE POIS3D(U,F,N,M,L,A,B,C,NMIN,MMIN,LMIN,
     1                  MAXITE,TOL,RESIDU,IREL)
C***************************************************************
C  .Parallel MG solver for 3D Poisson equation with Dirichlet BC
C   on the domain (0,A)x(0,B)x(0,C):
C
C           - LAPLACE(U)=F
C
C  .Arguments:
C
C   U      : Real array. The size of U should be at least
C               (IE1-IS1+1)*(JE1-JS1+1)*(KE1-KS1+1) +
C               (IE2-IS2+1)*(JE2-JS2+1)*(KE2-KS2+1) + ...
C            The first (IE1-IS1+1)*(JE1-JS1+1)*(KE1-KS1+1)
C            elements contain the initial solution and boundary
C            conditions on the finest grid.
C            (where ISi=ISTART(i), IEi=IEND(i), etc.).
C            (Input & Output).
C   F      : Real array. The size of F should be at least
C               (IE1-IS1+1)*(JE1-JS1+1)*(KE1-KS1+1)+
C               (IE2-IS2+1)*(JE2-JS2+1)*(KE2-KS2+1)+...
C            The first (IE1-IS1+1)*(JE1-JS1+1)*(KE1-KS1+1)
C            elements contain the rhs of the equation
C            (Input, only the values of F on the inner grid
C            points are used).
C   N,M,L  : Grid dimensions. The grid steps HX, HY and HZ are
C            given by:
C                        HX=A/N, HY=B/M, HZ=C/L
C            (Input only).
C   NMIN,MMIN,LMIN : Dimensions of the coarsest grid (Input).
C   MAXITE : Maxi. # of MG cycles to be performed on the finest
C            grid (Input).
C   TOL    : Convergence criteria on the residual (Input).
C   RESIDU : The final residual of the approximate solution
C            (Output only).
C   IREL   : Specify type of relaxation to be used.
C            Currently ignored. (red-black point-wise GS
C            relaxation is always used).
C            (Input)
C***************************************************************
C
C Warning: The switching between coarse and fine grids should be
C          the same on all tasks. It depends (especially when
C          IVCYCLE.ne.0) on the residual returned by MULTIG.
C          Since the residual checking on different machine may
C          yield different results on a heteogeneous system,
C          we use a synchronization mechanism to ensure that the
C          same grid switching procedure is used on all tasks in
C          which the master task decides the grid switching
C          strategy and broadcasts its decision to child tasks.
C
C The logical function CHKCOND is used to synchronize condition
C checking between tasks (especially for comparing floating
C point numbers).
      INCLUDE 'pvmp3d.cmm'
      LOGICAL CHKCOND
C
C MAXITG is the maxi. # of relaxations performed on the
C coarsest grid.
      PARAMETER (MAXITG=40)
      DIMENSION U(*),F(*)
C
C Local arrays:
C    MIT  = Maxi. # of MG cycles on coarser grids.
C    NIT  = # of MG cycles performed on the current grid.
C    TOL1 = Convergence criteria on the current grid.
C           (obtained by multiplying the residual of the
C           immediate finer grid by FACTOR).
      DIMENSION MIT(20),NIT(20),TOL1(20)
C
      CHARACTER*50 SPACES,DASHES
      DATA SPACES/'                                           '/
      DATA DASHES/'-------------------------------------------'/
      DATA FACTOR/0.1/
C
C Initialize counters for number of floating point operations
      CNTADD=0.0D0
      CNTMUL=0.0D0
C
C Perform MAXITE relaxation sweeps when there is only
C one grid level.
      IF (MOD(N,2).NE.0.OR.MOD(M,2).NE.0.OR.MOD(L,2).NE.0
     1    .OR.N.LE.NMIN.OR.M.LE.MMIN.OR.L.LE.LMIN) THEN
         DO 10 I=1,MAXITE
            CALL MULTIG(U,F,N,M,L,A,B,C,IREL,ISTART(1),IEND(1),
     1            JSTART(1),JEND(1),KSTART(1),KEND(1))
            RESIDU=EVALRES(U,F,N,M,L,A,B,C,ISTART(1),IEND(1),
     1            JSTART(1),JEND(1),KSTART(1),KEND(1))
            IF (ME.EQ.0) PRINT 1000,SPACES(1:1*2),N,M,L,
     1                             DASHES(1:(10-1)*2+3),I,RESIDU
            IF (CHKCOND(RESIDU.LT.TOL)) RETURN
10       CONTINUE
         RETURN
      ENDIF
C
C--- Check if we should use V-cycle
      IF (IVCYCLE.NE.0) GOTO 20000
C
*+++++++++++++++++++++++ Variable MG cycle +++++++++++++++++++++
10000 CONTINUE
C
C init ERR0, ERR1 and RESIDU to an arbitrary nonzero value
C (they are referenced but never initialized on child tasks)
      ERR0=1.0
      ERR1=1.0
      RESIDU=1.0
C
C Final residual requirement on the finest grid.
      TOL1(1)=TOL
C
C Set maxi. # of MG cycles for all grids.
      DO 1 I=2,10
         MIT(I)=2
1     CONTINUE
      MIT(1)=MAXITE
C
C Initialize control variables.
C LEVEL=current grid level. (1=finest)
C N0,M0,L0=dimensions of the current grid.
C N1=entry position in U & F corresponding to the current grid.
      LEVEL=1
      IS=ISTART(LEVEL)
      IE=IEND(LEVEL)
      JS=JSTART(LEVEL)
      JE=JEND(LEVEL)
      KS=KSTART(LEVEL)
      KE=KEND(LEVEL)
      NIT(LEVEL)=0
      N0=N
      M0=M
      L0=L
      N1=1
C
C Main loop
100   CONTINUE
C
C NBRIT=# of relaxation sweeps performed on the current grid.
      NBRIT=0
C
C Compute the exact solution by 1 relaxation if N0=M0=L0=2
      IF (N0.EQ.2.AND.M0.EQ.2.AND.L0.EQ.2) THEN
         NBRIT=1
         CALL MULTIG(U(N1),F(N1),N0,M0,L0,A,B,C,IREL,
     1               IS,IE,JS,JE,KS,KE)
         ERR1=EVALRES(U(N1),F(N1),N0,M0,L0,A,B,C,
     1                IS,IE,JS,JE,KS,KE)
         GOTO 3
      END IF
C
C Set initial residual to infinity.
      IF (ME.EQ.0) ERR1=1.0E20
C
C Relaxation loop
2     IF (ME.EQ.0) ERR0=ERR1
      CALL MULTIG(U(N1),F(N1),N0,M0,L0,A,B,C,IREL,
     1            IS,IE,JS,JE,KS,KE)
      ERR1=EVALRES(U(N1),F(N1),N0,M0,L0,A,B,C,IS,IE,JS,JE,KS,KE)
      NBRIT=NBRIT+1
C
C Check if this is the coarsest grid.
      IF (MOD(N0,2).NE.0.OR.MOD(M0,2).NE.0.OR.MOD(L0,2).NE.0
     1    .OR.N0.LE.NMIN.OR.M0.LE.MMIN.OR.L0.LE.LMIN) THEN
C
C Yes ===> Continue relaxation until the convergence requirement
C          is satisfied or the # of relaxations exceeds MAXITG.
         IF ( CHKCOND(NBRIT.LT.MAXITG.AND.ERR1.GE.TOL1(LEVEL)) )
     1   THEN
            GOTO 2
         ELSE
            GOTO 3
         END IF
      END IF
C
C No ===> Continue the relaxation while the convergence factor
C         is not greater than 0.25 and the convergence
C         requirement is not yet satisfied.
      IF ( CHKCOND(ERR1/ERR0.LE.0.25.AND.ERR1.GE.TOL1(LEVEL)) )
     1  GOTO 2
C
3     CONTINUE
C
C Print the trace of MG loop.
***      IF (ME.EQ.0 .AND. LEVEL.EQ.1) THEN
      IF (ME.EQ.0) THEN
         PRINT 1000,SPACES(1:LEVEL*2),N0,M0,L0,
     1              DASHES(1:(10-LEVEL)*2+3),NIT(1),ERR1
      ENDIF
C
C If the convergence requirement is satisfied or the current
C grid is the coarsest one, then goto next finer grid. Otherwise
C goto next coarser grid.
      IF ( CHKCOND( (ERR1.LT.TOL1(LEVEL)) .OR.
     1              (NIT(LEVEL).GE.MIT(LEVEL)) .OR.
     2              (MOD(N0,2).NE.0) .OR. (N0.EQ.2) .OR.
     3              (MOD(M0,2).NE.0) .OR. (M0.EQ.2) .OR.
     4              (MOD(L0,2).NE.0) .OR. (L0.EQ.2) .OR.
     5              N0.LE.NMIN .OR. M0.LE.MMIN .OR. L0.LE.LMIN )
     6   ) THEN
C
C------------------ Go up -------------------------
C
C End of the MG task if the current grid is the finest one.
         IF ( CHKCOND( LEVEL.EQ.1 .AND.
     1                 ( ERR1.LT.TOL1(1) .OR.
     2                   NIT(1).GE.MIT(1)) ) ) THEN
            RESIDU=ERR1
            RETURN
         END IF
         LEVEL=LEVEL-1
         N0=N0+N0
         M0=M0+M0
         L0=L0+L0
         IS1=ISTART(LEVEL)
         IE1=IEND(LEVEL)
         JS1=JSTART(LEVEL)
         JE1=JEND(LEVEL)
         KS1=KSTART(LEVEL)
         KE1=KEND(LEVEL)
         N2=N1-(IE1-IS1+1)*(JE1-JS1+1)*(KE1-KS1+1)
C
C Coarse grid correction
         CALL MULTPR(U(N1),U(N2),N0,M0,L0,IS,IE,JS,JE,KS,KE,
     1               IS1,IE1,JS1,JE1,KS1,KE1)
         N1=N2
         IS=IS1
         IE=IE1
         JS=JS1
         JE=JE1
         KS=KS1
         KE=KE1
         GOTO 100
      ELSE
C
C-------------- Go down ----------------------------
C
         NIT(LEVEL)=NIT(LEVEL)+1
         LEVEL=LEVEL+1
         TOL1(LEVEL)=ERR1*FACTOR
         NIT(LEVEL)=0
         N2=N1+(IE-IS+1)*(JE-JS+1)*(KE-KS+1)
         IS1=ISTART(LEVEL)
         IE1=IEND(LEVEL)
         JS1=JSTART(LEVEL)
         JE1=JEND(LEVEL)
         KS1=KSTART(LEVEL)
         KE1=KEND(LEVEL)
C
C Residual transfer
         CALL MULTRE(U(N1),F(N1),U(N2),F(N2),N0,M0,L0,A,B,C,
     1               IS,IE,JS,JE,KS,KE,IS1,IE1,JS1,JE1,KS1,KE1)
         N0=N0/2
         M0=M0/2
         L0=L0/2
         N1=N2
         IS=IS1
         IE=IE1
         JS=JS1
         JE=JE1
         KS=KS1
         KE=KE1
         GOTO 100
      END IF
C
C End of the MG loop with variable cycle
*++++++++++++++++++++++++++++ V cycle ++++++++++++++++++++++++++
20000 CONTINUE
C
      MGCOUNT=0
      LEVEL=1
      IS=ISTART(LEVEL)
      IE=IEND(LEVEL)
      JS=JSTART(LEVEL)
      JE=JEND(LEVEL)
      KS=KSTART(LEVEL)
      KE=KEND(LEVEL)
      N0=N
      M0=M
      L0=L
      N1=1
C
C Main loop
200   CONTINUE
C
C If not finest grid, then perform NP relaxation sweeps and
C transfer the residual to next coarser grid.
      IF ( .NOT. ( MOD(N0,2).NE.0 .OR.
     1             MOD(M0,2).NE.0 .OR.
     2             MOD(L0,2).NE.0 .OR.
     3             N0.LE.NMIN .OR.
     4             M0.LE.MMIN .OR.
     5             L0.LE.LMIN ) ) THEN
         DO 201 I=1,NP
            CALL MULTIG(U(N1),F(N1),N0,M0,L0,A,B,C,IREL,
     1                  IS,IE,JS,JE,KS,KE)
201      CONTINUE
         LEVEL=LEVEL+1
         N2=N1+(IE-IS+1)*(JE-JS+1)*(KE-KS+1)
         IS1=ISTART(LEVEL)
         IE1=IEND(LEVEL)
         JS1=JSTART(LEVEL)
         JE1=JEND(LEVEL)
         KS1=KSTART(LEVEL)
         KE1=KEND(LEVEL)
C
C Residual transfer
         CALL MULTRE(U(N1),F(N1),U(N2),F(N2),N0,M0,L0,A,B,C,
     1               IS,IE,JS,JE,KS,KE,IS1,IE1,JS1,JE1,KS1,KE1)
         N0=N0/2
         M0=M0/2
         L0=L0/2
         N1=N2
         IS=IS1
         IE=IE1
         JS=JS1
         JE=JE1
         KS=KS1
         KE=KE1
         GOTO 200
      ENDIF
C
C Solution on the coarsest grid (we just perform a sufficient
C number of relaxation sweeps to reduce the residual by a
C given factor
      IF (N0.EQ.2 .AND. M0.EQ.2 .AND. L0.EQ.2) THEN
C One relaxation sweep solves the problem if
C the coarsest grid = 2x2x2
         CALL MULTIG(U(N1),F(N1),N0,M0,L0,A,B,C,IREL,
     1               IS,IE,JS,JE,KS,KE)
      ELSE
         TOL0=0.01*EVALRES(U(N1),F(N1),N0,M0,L0,A,B,C,
     1                     IS,IE,JS,JE,KS,KE)
         I=0
300      CONTINUE
         CALL MULTIG(U(N1),F(N1),N0,M0,L0,A,B,C,IREL,
     1               IS,IE,JS,JE,KS,KE)
         RESIDU=EVALRES(U(N1),F(N1),N0,M0,L0,A,B,C,
     1                  IS,IE,JS,JE,KS,KE)
         I=I+1
         IF ( CHKCOND( RESIDU.GT.TOL0 .AND. I.LT.MAXITG ) )
     1      GOTO 300
      ENDIF
C
C If not finest grid then prolongation & coarse grid correction
400   CONTINUE
      IF (LEVEL.EQ.1) THEN
         MGCOUNT=MGCOUNT+1
         RESIDU=EVALRES(U(N1),F(N1),N0,M0,L0,A,B,C,
     1                  IS,IE,JS,JE,KS,KE)
         IF (ME .EQ. 0) THEN
            PRINT 1000,SPACES(1:LEVEL*2),N0,M0,L0,
     1                 DASHES(1:(10-LEVEL)*2+3),
     2                 MGCOUNT,RESIDU
         ENDIF
         IF ( CHKCOND( MGCOUNT.LT.MAXITE .AND. RESIDU.GT.TOL ) )
     1      GOTO 200
         RETURN
      ENDIF
C
C prolongation and coarse grid correction
         LEVEL=LEVEL-1
         N0=N0+N0
         M0=M0+M0
         L0=L0+L0
         IS1=ISTART(LEVEL)
         IE1=IEND(LEVEL)
         JS1=JSTART(LEVEL)
         JE1=JEND(LEVEL)
         KS1=KSTART(LEVEL)
         KE1=KEND(LEVEL)
         N2=N1-(IE1-IS1+1)*(JE1-JS1+1)*(KE1-KS1+1)
C
C Coarse grid correction
         CALL MULTPR(U(N1),U(N2),N0,M0,L0,IS,IE,JS,JE,KS,KE,
     1               IS1,IE1,JS1,JE1,KS1,KE1)
         N1=N2
         IS=IS1
         IE=IE1
         JS=JS1
         JE=JE1
         KS=KS1
         KE=KE1
C
C Post relaxation
      DO 202 I=1,NQ
         CALL MULTIG(U(N1),F(N1),N0,M0,L0,A,B,C,IREL,
     1               IS,IE,JS,JE,KS,KE)
202   CONTINUE
      GOTO 400
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
1000  FORMAT(1X,A,I4,1X,I4,1x,I4,A,'it #: ',I3,
     1       '      Res=',1PG15.8)
      END
C
C...............................................................
C
      LOGICAL FUNCTION CHKCOND(L)
      INCLUDE 'pvmp3d.cmm'
      LOGICAL L
C
C Synchronize condition checking.
C . The master task checks the condition L and broadcasts the
C   result to child tasks.
C . The child tasks receive the result from the master task
C
      IF (NPROC.LE.1) THEN
         CHKCOND=L
         RETURN
      ENDIF
C
      IF (ME.EQ.0) THEN
         IF (L) THEN
            ICOND=1
         ELSE
            ICOND=0
         ENDIF
         CALL PVMFINITSEND(PVMDEFAULT,INFO)
         IF (INFO.LT.0) CALL EXIT(49)
         CALL PVMFPACK(INTEGER4,ICOND,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(50)
         MSGTAG=7
         CALL PVMFMCAST(NPROC-1,TIDS(1),MSGTAG,INFO)
         IF (INFO.LT.0) CALL EXIT(51)
      ELSE
         MSGTAG=7
         CALL PVMFRECV(PARENTID,MSGTAG,INFO)
         IF (INFO.LT.0) CALL EXIT(52)
         CALL PVMFUNPACK(INTEGER4,ICOND,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(53)
      ENDIF
      CHKCOND=(ICOND.NE.0)
      RETURN
      END
C
C...............................................................
C
      SUBROUTINE MULTIG(U,F,N,M,L,A,B,C,IREL,IS,IE,JS,JE,KS,KE)
C
C Perform one sweep of point-wise or line-wise GS relaxation
C (depending on the value of IREL).
C
C !!! The value of IREL is ignored and red-black point-wise
C     GS relaxation is used.
C
C THIS SUBROUTINE IS NOT OPTIMAL. Some arithmetic operations can
C be saved by exploiting the properties of RED-BLACK
C Gauss-Seidel iteration.
C
      INCLUDE 'pvmp3d.cmm'
      DIMENSION U(IS:IE,JS:JE,KS:KE),F(IS:IE,JS:JE,KS:KE)
C
      HX2=(N/A)**2
      HY2=(M/B)**2
      HZ2=(L/C)**2
      H2=2.0*(HX2+HY2+HZ2)
      ONEDH2=1.0/H2
C
C red-black point-wise GS relaxation
C (!!! do not change the ordering below, it must be consistent
C  with the restriction and prolongation routines).
C
C--- Relaxation on red points (I+J+K EVEN)
      DO 10 K=KS+1,KE-1
      DO 10 J=JS+1,JE-1
      DO 10 I=IS+2-MOD(K+J+IS,2),IE-1,2
         U(I,J,K)=(F(I,J,K)+HX2*(U(I+1,J,K)+U(I-1,J,K))
     1                     +HY2*(U(I,J+1,K)+U(I,J-1,K))
     2                     +HZ2*(U(I,J,K+1)+U(I,J,K-1)))*ONEDH2
10    CONTINUE
C
C Exchange new values of U with other tasks (on red points)
      MSGTAG=3
      CALL XCHGDATA(U,N,M,L,IS,IE,JS,JE,KS,KE,MSGTAG,2)
C
C--- Relaxation on black points (I+J+K ODD)
      DO 11 K=KS+1,KE-1
      DO 11 J=JS+1,JE-1
      DO 11 I=IS+MOD(K+J+IS,2)+1,IE-1,2
         U(I,J,K)=(F(I,J,K)+HX2*(U(I+1,J,K)+U(I-1,J,K))
     1                     +HY2*(U(I,J+1,K)+U(I,J-1,K))
     2                     +HZ2*(U(I,J,K+1)+U(I,J,K-1)))*ONEDH2
11    CONTINUE
C
C Exchange new values of U with other tasks (on black points)
      MSGTAG=4
      CALL XCHGDATA(U,N,M,L,IS,IE,JS,JE,KS,KE,MSGTAG,1)
C
      CALL TIMING(1,N,M,L)
C
C Count # of floating point operations of the relaxation sweep
C (!!! The # of operations may be largely reduced if HX==HY==HZ)
      CNTTMP=(N-1)*DBLE(M-1)*(L-1)
      CNTADD=CNTADD+6.0D0*CNTTMP
      CNTMUL=CNTMUL+4.0D0*CNTTMP
C
      RETURN
      END
C
C...............................................................
C
      FUNCTION EVALRES(U,F,N,M,L,A,B,C,IS,IE,JS,JE,KS,KE)
      INCLUDE 'pvmp3d.cmm'
      DIMENSION U(IS:IE,JS:JE,KS:KE),F(IS:IE,JS:JE,KS:KE)
C
      HX2=(N/A)**2
      HY2=(M/B)**2
      HZ2=(L/C)**2
      H2=2.0*(HX2+HY2+HZ2)
C
C Compute the residual (L^2 norm), the residual on black points
C are zero since they have just been relaxed.
      RESIDU=0.0
      DO 20 K=KS+1,KE-1
      DO 20 J=JS+1,JE-1
      DO 20 I=IS+2-MOD(K+J+IS,2),IE-1,2
         R=F(I,J,K)+HX2*(U(I+1,J,K)+U(I-1,J,K))
     1             +HY2*(U(I,J+1,K)+U(I,J-1,K))
     2             +HZ2*(U(I,J,K+1)+U(I,J,K-1))
     3             -H2*U(I,J,K)
         RESIDU=RESIDU+R*R
20    CONTINUE
C
C Compute the total of residuals from all tasks.
      IF (ME.EQ.0) THEN
C
C Collect residuals from all child-tasks
         IF (NPROC.GT.1) THEN
            MSGTAG=1
            DO 30 I=1,NPROC-1
               CALL PVMFRECV(-1,MSGTAG,INFO)
               IF (INFO.LT.0) CALL EXIT(54)
               CALL PVMFUNPACK(NREALTYPE,R,1,1,INFO)
               IF (INFO.LT.0) CALL EXIT(55)
               RESIDU=RESIDU+R
30          CONTINUE
         ENDIF
         EVALRES=SQRT(RESIDU/(HX2*HY2*HZ2))
C
C Broadcast total residual to all child-tasks
CCC         IF (NPROC.GT.1) THEN
CCC            CALL PVMFINITSEND(PVMDEFAULT,INFO)
CCC            IF (INFO.LT.0) CALL EXIT(56)
CCC            CALL PVMFPACK(NREALTYPE,RESIDU,1,1,INFO)
CCC            IF (INFO.LT.0) CALL EXIT(57)
CCC            MSGTAG=2
CCC            CALL PVMFMCAST(NPROC-1,TIDS(1),MSGTAG,INFO)
CCC            IF (INFO.LT.0) CALL EXIT(58)
CCC         ENDIF
      ELSE
C
C Send my part of residual to the master task
         CALL PVMFINITSEND(PVMDEFAULT,INFO)
         IF (INFO.LT.0) CALL EXIT(59)
         CALL PVMFPACK(NREALTYPE,RESIDU,1,1,INFO)
         IF (INFO.LT.0) CALL EXIT(60)
         MSGTAG=1
         CALL PVMFSEND(PARENTID,MSGTAG,INFO)
         IF (INFO.LT.0) CALL EXIT(61)
C
C Get total residual from the master task
CCC         MSGTAG=2
CCC         CALL PVMFRECV(PARENTID,MSGTAG,INFO)
CCC         IF (INFO.LT.0) CALL EXIT(62)
CCC         CALL PVMFUNPACK(NREALTYPE,RESIDU,1,1,INFO)
CCC         IF (INFO.LT.0) CALL EXIT(63)
      ENDIF
C
C # of operations for the residual evaluation
      CNTTMP=(N-1)*DBLE(M-1)*(L-1)
      CNTADD=CNTADD+8.0D0*0.5*CNTTMP
      CNTMUL=CNTMUL+5.0D0*0.5*CNTTMP
C
      RETURN
      END
C
C...............................................................
C
      SUBROUTINE MULTRE(U,F,V,G,N,M,L,A,B,C,IFS,IFE,JFS,JFE,
     1                  KFS,KFE,IGS,IGE,JGS,JGE,KGS,KGE)
C
C Residual transfer by injection.
C
C  U,F ===> Fine grid functions
C  V,G ===> Coarse grid functions
C
      INCLUDE 'pvmp3d.cmm'
      DIMENSION U(IFS:IFE,JFS:JFE,KFS:KFE),
     1          F(IFS:IFE,JFS:JFE,KFS:KFE),
     2          V(IGS:IGE,JGS:JGE,KGS:KGE),
     3          G(IGS:IGE,JGS:JGE,KGS:KGE)
C R is used for full weighted residual transfer
***      DIMENSION R(-1:1,-1:1,-1:1)
      DATA ZERO/0.0/,HALF/0.5/
C
      N1=N/2
      M1=M/2
      L1=L/2
      HX2=(N/A)**2
      HY2=(M/B)**2
      HZ2=(L/C)**2
      H2=2.0*(HX2+HY2+HZ2)
C
C Residual transfer. Residual weighting is necessary because we
C use Red-Black ordering in the relaxation.
C
C Partial weighting:
C Since,
C     RC(IC,JC,KC)=( RF(IF-1,JF,KF) + RF(IF+1,JF,KF) +
C                    RF(IF,JF-1,KF) + RF(IF,JF+1,KF) +
C                    RF(IF,JF,KF-1) + RF(IF,JF,KF+1) +
C                    RF(IF,JF,KF)*6.0 )/12.0
C where IF=IC+IC, JF=JC+JC and KF=KC+KC.
C All terms except the last one in the above expression are null
C (they are black points, i.e., I+J+K odd), so we have:
C     RC(IC,JC,KC) = RF(IF,JF,KF)*6.0/12.0 = RF(IF,JF,KF)/2.0
C
C Full weighting:
C     RC(IC,JC,KC)=( Sum for (-1<=i<=1, -1<=j<=1, -1<=k<=1 )
C                    c(i,j,k)*RF(IF+i,JF+j,KF+k) ) / 64
C where c(i,j,k)=8/2**(abs(i)+abs(j)+abs(k)).
C Note: we have RF(IF+i,JF+j,KF+k)=0 if i+j+k odd, due to the
C       relaxation ordering.
C
C The convergence rate with full weighting is faster than with
C partial weighting, but the overall efficiency is worse in
C many cases.
C
      DO 1 K=KGS+1,KGE-1
         KK=K+K
         DO 2 J=JGS+1,JGE-1
            JJ=J+J
            DO 3 I=IGS+1,IGE-1
               II=I+I
C------------- Partial weighting for resdual transfer. ---------
               G(I,J,K)=( F(II,JJ,KK)
     1                   +HX2*(U(II-1,JJ,KK)+U(II+1,JJ,KK))
     2                   +HY2*(U(II,JJ-1,KK)+U(II,JJ+1,KK))
     3                   +HZ2*(U(II,JJ,KK-1)+U(II,JJ,KK+1))
     4                   -H2*U(II,JJ,KK)
     5                  )*HALF
C---------------------------------------------------------------
3           CONTINUE
2        CONTINUE
1     CONTINUE
C
C Initial solution on the coarse grid = 0.
C
      DO 10 K=KGS,KGE
      DO 10 J=JGS,JGE
      DO 10 I=IGS,IGE
         V(I,J,K)=ZERO
10    CONTINUE
C
C # of operations
      CNTTMP=(N1-1)*DBLE(M1-1)*(L1-1)
      CNTADD=CNTADD+7.0D0*CNTTMP
      CNTMUL=CNTMUL+5.0D0*CNTTMP
C
      RETURN
      END
C
C...............................................................
C
      SUBROUTINE MULTPR(V,U,N,M,L,IGS,IGE,JGS,JGE,KGS,KGE,
     1                  IFS,IFE,JFS,JFE,KFS,KFE)
C
C Coarse grid correction by linear interpolation.
C We only have to interpolate on black points (I+J+K odd).
C
      INCLUDE 'pvmp3d.cmm'
      DIMENSION V(IGS:IGE,JGS:JGE,KGS:KGE),
     1          U(IFS:IFE,JFS:JFE,KFS:KFE)
      DATA HALF/0.5/, EIGHTH/0.125/
***      DATA QUART/0.25/
C
      N1=N/2
      M1=M/2
      L1=L/2
CKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKK
      DO 10 K=KGS+1,KGE-1
         KK=K+K
CJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJ
         DO 11 J=JGS+1,JGE-1
            JJ=J+J
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
C KK even, JJ even, II even
***            DO 12 I=IGS+1,IGE-1
***               II=I+I
***               U(II,JJ,KK)=U(II,JJ,KK)+V(I,J,K)
***12          CONTINUE
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
C KK even, JJ even, II odd
            DO 13 I=IGS,IGE-1
               II=I+I+1
               U(II,JJ,KK)=U(II,JJ,KK)+HALF*(V(I,J,K)+V(I+1,J,K))
13          CONTINUE
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
11       CONTINUE
CJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJ
         DO 14 J=JGS,JGE-1
            JJ=J+J+1
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
C KK even, JJ odd, II even
            DO 15 I=IGS+1,IGE-1
               II=I+I
               U(II,JJ,KK)=U(II,JJ,KK)+HALF*(V(I,J,K)+V(I,J+1,K))
15          CONTINUE
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
C KK even, JJ odd, II odd
***            DO 16 I=IGS,IGE-1
***               II=I+I+1
***               U(II,JJ,KK)=U(II,JJ,KK)+
***     1                     QUART*(V(I,J,K)+V(I+1,J,K)+
***     2                            V(I,J+1,K)+V(I+1,J+1,K))
16          CONTINUE
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
14       CONTINUE
CJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJ
10    CONTINUE
CKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKK
      DO 20 K=KGS,KGE-1
         KK=K+K+1
CJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJ
         DO 21 J=JGS+1,JGE-1
            JJ=J+J
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
C KK odd, JJ even, II even
            DO 22 I=IGS+1,IGE-1
               II=I+I
               U(II,JJ,KK)=U(II,JJ,KK)+HALF*(V(I,J,K)+V(I,J,K+1))
22          CONTINUE
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
C KK odd, JJ even, II odd
***            DO 23 I=IGS,IGE-1
***               II=I+I+1
***               U(II,JJ,KK)=U(II,JJ,KK)+
***     1                     QUART*(V(I,J,K)+V(I+1,J,K)+
***     2                            V(I,J,K+1)+V(I+1,J,K+1))
23          CONTINUE
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
21       CONTINUE
CJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJ
         DO 24 J=JGS,JGE-1
            JJ=J+J+1
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
C KK odd, JJ odd, II even
***            DO 25 I=IGS+1,IGE-1
***               II=I+I
***               U(II,JJ,KK)=U(II,JJ,KK)+
***     1                     QUART*(V(I,J,K)+V(I,J+1,K)+
***     2                            V(I,J,K+1)+V(I,J+1,K+1))
25          CONTINUE
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
C KK odd, JJ odd, II odd
            DO 26 I=IGS,IGE-1
               II=I+I+1
               U(II,JJ,KK)=U(II,JJ,KK)+
     1                     EIGHTH*( V(I,J,K) + V(I+1,J,K) +
     3                              V(I,J+1,K) + V(I+1,J+1,K) +
     4                              V(I,J,K+1) + V(I+1,J,K+1) +
     5                              V(I,J+1,K+1) + V(I+1,J+1,K+1)
     6                            )
26          CONTINUE
CIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
24       CONTINUE
CJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJ
20    CONTINUE
CKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKK
C
C Exchange data with other tasks (on black points)
      MSGTAG=5
      CALL XCHGDATA(U,N,M,L,IFS,IFE,JFS,JFE,KFS,KFE,MSGTAG,1)
C
C # of operations
      CNTADD=CNTADD + 2.0D0*(
     1                       (L1-1)*DBLE(M1-1)*(N1-0) +
     2                       (L1-1)*DBLE(M1-0)*(N1-1) +
     3                       (L1-0)*DBLE(M1-1)*(N1-1)
     4                      )
     5              + 8.0D0*(L1-0)*DBLE(M1-0)*(N1-0)
      CNTMUL=CNTMUL + (L1-1)*DBLE(M1-1)*(N1-0)
     1              + (L1-1)*DBLE(M1-0)*(N1-1)
     2              + (L1-0)*DBLE(M1-1)*(N1-1)
     3              + (L1-0)*DBLE(M1-0)*(N1-0)
C
      RETURN
      END
C
C---------------------------------------------------------------
C
      SUBROUTINE XCHGDATA(U,N,M,L,IS,IE,JS,JE,KS,KE,MSGTAG,NRB)
      INCLUDE 'pvmp3d.cmm'
      DIMENSION U(IS:IE,JS:JE,KS:KE)
      LOGICAL SENT,RECVED,PARITY
C
C Exchange boundary data with the neibours
C     NRB.eq.0:  all points
C     NRB.eq.1:  for I+J+K odd (black points)
C     NRB.eq.2:  for I+J+K even (red points)
C
      INCI=1
      INCJ=INCI*(IE-IS+1)
      INCK=INCJ*(JE-JS+1)
C
      CALL TIMING(1,N,M,L)
C
C ========================= X Surfaces =========================
      SENT=.FALSE.
      RECVED=.FALSE.
      PARITY= ( MOD(MEX,2).EQ.0 )
      IF (PARITY .OR. IE.LE.IS+1.OR.JE.LE.JS+1.OR.KE.LE.KS+1)
     1   GOTO 200
100   CONTINUE
C
C Send to previous
      CALL XCHGD0( U(IS+1,JS+1,KS+1),
     1             INCJ,JS+1,JE-1,M,
     2             INCK,KS+1,KE-1,L,
     3             IS+1,N,
     4             MSGTAG, NRB, 1,
     5             MEX-1,MEY,MEZ )
C Send to next
      CALL XCHGD0( U(IE-1,JS+1,KS+1),
     1             INCJ,JS+1,JE-1,M,
     2             INCK,KS+1,KE-1,L,
     3             IE-1,N,
     4             MSGTAG, NRB, 1,
     5             MEX+1,MEY,MEZ )
      SENT=.TRUE.
      IF (RECVED) GOTO 300
200   CONTINUE
C
C Receive from previous
      CALL XCHGD0( U(IS,JS+1,KS+1),
     1             INCJ,JS+1,JE-1,M,
     2             INCK,KS+1,KE-1,L,
     3             IS,N,
     4             MSGTAG, NRB, 0,
     5             MEX-1,MEY,MEZ )
C Receive from next
      CALL XCHGD0( U(IE,JS+1,KS+1),
     1             INCJ,JS+1,JE-1,M,
     2             INCK,KS+1,KE-1,L,
     3             IE,N,
     4             MSGTAG, NRB, 0,
     5             MEX+1,MEY,MEZ )
      RECVED=.TRUE.
      IF (.NOT. SENT) GOTO 100
C
300   CONTINUE
C
C =================== Y Surfaces + X Edges =====================
      SENT=.FALSE.
      RECVED=.FALSE.
      PARITY= ( MOD(MEY,2).EQ.0 )
      IF (PARITY .OR. IE.LE.IS+1.OR.JE.LE.JS+1.OR.KE.LE.KS+1)
     1   GOTO 201
101   CONTINUE
C
C Send to previous
      CALL XCHGD0( U(IS,JS+1,KS+1),
     1             INCI,IS,IE,N,
     2             INCK,KS+1,KE-1,L,
     3             JS+1,M,
     4             MSGTAG, NRB, 1,
     5             MEX,MEY-1,MEZ )
C Send to next
      CALL XCHGD0( U(IS,JE-1,KS+1),
     1             INCI,IS,IE,N,
     2             INCK,KS+1,KE-1,L,
     3             JE-1,M,
     4             MSGTAG, NRB, 1,
     5             MEX,MEY+1,MEZ )
      SENT=.TRUE.
      IF (RECVED) GOTO 301
201   CONTINUE
C
C Receive from previous
      CALL XCHGD0( U(IS,JS,KS+1),
     1             INCI,IS,IE,N,
     2             INCK,KS+1,KE-1,L,
     3             JS,M,
     4             MSGTAG, NRB, 0,
     5             MEX,MEY-1,MEZ )
C Receive from next
      CALL XCHGD0( U(IS,JE,KS+1),
     1             INCI,IS,IE,N,
     2             INCK,KS+1,KE-1,L,
     3             JE,M,
     4             MSGTAG, NRB, 0,
     5             MEX,MEY+1,MEZ )
      RECVED=.TRUE.
      IF (.NOT. SENT) GOTO 101
C
301   CONTINUE
C
C =================== Z Surfaces + XY Edges ====================
      SENT=.FALSE.
      RECVED=.FALSE.
      PARITY= ( MOD(MEZ,2).EQ.0 )
      IF (PARITY .OR. IE.LE.IS+1.OR.JE.LE.JS+1.OR.KE.LE.KS+1)
     1   GOTO 202
102   CONTINUE
C
C Send to previous
      CALL XCHGD0( U(IS,JS,KS+1),
     1             INCI,IS,IE,N,
     2             INCJ,JS,JE,M,
     3             KS+1,L,
     4             MSGTAG, NRB, 1,
     5             MEX,MEY,MEZ-1 )
C Send to next
      CALL XCHGD0( U(IS,JS,KE-1),
     1             INCI,IS,IE,N,
     2             INCJ,JS,JE,M,
     3             KE-1,L,
     4             MSGTAG, NRB, 1,
     5             MEX,MEY,MEZ+1 )
      SENT=.TRUE.
      IF (RECVED) GOTO 302
202   CONTINUE
C
C Receive from previous
      CALL XCHGD0( U(IS,JS,KS),
     1             INCI,IS,IE,N,
     2             INCJ,JS,JE,M,
     3             KS,L,
     4             MSGTAG, NRB, 0,
     5             MEX,MEY,MEZ-1 )
C Receive from next
      CALL XCHGD0( U(IS,JS,KE),
     1             INCI,IS,IE,N,
     2             INCJ,JS,JE,M,
     3             KE,L,
     4             MSGTAG, NRB, 0,
     5             MEX,MEY,MEZ+1 )
      RECVED=.TRUE.
      IF (.NOT. SENT) GOTO 102
C
302   CONTINUE
C
      CALL TIMING(2,N,M,L)
C
      RETURN
      END
C
C---------------------------------------------------------------
C
      SUBROUTINE XCHGD0( U,
     1                   INCI,IS,IE,N,
     2                   INCJ,JS,JE,M,
     3                   K,L,
     4                   MSGTAG, NRB, ISND,
     5                   MEX1,MEY1,MEZ1 )
      INCLUDE 'pvmp3d.cmm'
      DIMENSION U(0:*)
      INTEGER TID
      LOGICAL INIFLG
C
C This subroutine sends to or receives from the task
C (MEX1,MEY1,MEZ1) the following data:
C
C   ( (U(i*INCI+j*INCJ), i=IS,IE), j=JS,JE)
C
C   ISND.EQ.0 : Receiving data
C   ISND.NE.0 : Sending  data
C
C   NRB==0:     all points
C   NRB==1:     black points (I+J+K odd)
C   NRB==2:     red points (I+J+K even)
C
      IF (MEX1.LT.0 .OR. MEX1.GE.NPROCX .OR.
     1    MEY1.LT.0 .OR. MEY1.GE.NPROCY .OR.
     2    MEZ1.LT.0 .OR. MEZ1.GE.NPROCZ .OR.
     3    K.EQ.0 .OR. K.EQ.L) RETURN
C
      TID=TIDS(MEX1+(MEY1+MEZ1*NPROCY)*NPROCX)
      IS1=IS
      IE1=IE
      IF (IS1.EQ.0) IS1=IS1+1
      IF (IE1.EQ.N) IE1=IE1-1
      JS1=JS
      JE1=JE
      IF (JS1.EQ.0) JS1=JS1+1
      IF (JE1.EQ.M) JE1=JE1-1
      IF (IS1.GT.IE1 .OR. JS1.GT.JE1) RETURN
C
      INIFLG=.FALSE.
      DO 1 J=JS1,JE1
         IF (NRB.EQ.0) THEN
            I=IS1
            LEN=IE1-IS1+1
            INC=INCI
         ELSE
            I=IS1
            IF (NRB.EQ.1) THEN
               IF (MOD(K+J+I,2).EQ.0) I=I+1
            ELSE
               IF (MOD(K+J+I,2).NE.0) I=I+1
            ENDIF
            LEN=(IE1-I+1+1)/2
            INC=2*INCI
         ENDIF
         IF (LEN.GT.0) THEN
            I0=(I-IS)*INCI + (J-JS)*INCJ
            IF (ISND.NE.0) THEN
               IF (.NOT. INIFLG) THEN
***                     print 1000,mytid,' sending to ',tid
                  CALL PVMFINITSEND(PVMDEFAULT,INFO)
                  IF (INFO.LT.0) CALL EXIT(67)
                  INIFLG=.TRUE.
               ENDIF
               CALL PVMFPACK(NREALTYPE,U(I0),LEN,INC,INFO)
               IF (INFO.LT.0) CALL EXIT(68)
            ELSE
               IF (.NOT. INIFLG) THEN
***                     print 1000,mytid,' receiving from ',tid
***1000                 format (1x,z8,a,z8)
                  CALL PVMFRECV(TID,MSGTAG,INFO)
                  IF (INFO.LT.0) CALL EXIT(69)
                  INIFLG=.TRUE.
               ENDIF
               CALL PVMFUNPACK(NREALTYPE,U(I0),LEN,INC,INFO)
               IF (INFO.LT.0) CALL EXIT(70)
            ENDIF
         ENDIF
1     CONTINUE
      IF (ISND.NE.0 .AND. INIFLG) THEN
         CALL PVMFSEND(TID,MSGTAG,INFO)
         IF (INFO.LT.0) CALL EXIT(71)
      ENDIF
C
      RETURN
      END
C
C---------------------------------------------------------------
C
      SUBROUTINE TIMING(IFLG,N,M,L)
      INCLUDE 'pvmp3d.cmm'
      DOUBLE PRECISION TM(3)
C
C IFLG.EQ.0:  Initialization, first call to this subroutine
C IFLG.EQ.1:  CPU
C IFLG.EQ.2:  Communication
C IFLG.EQ.-1: End of timing
      IF (ME.NE.0) RETURN
      CALL GETTIME(TM)
      IF (IFLG.EQ.0) THEN
         OPEN(17,FILE='pvmp3d.timing')
         WRITE(17,1001) N,M,L,NPROCX,NPROCY,NPROCZ,NROUTE,
     1                  NENCODING
      ENDIF
      WRITE(17,1000) IFLG,TM(1),TM(2),TM(3),N,M,L
      IF (IFLG.EQ.-1) CLOSE(17)
1000  FORMAT(1X,I2,1X,1PG15.8,1X,1PG15.8,1X,1PG22.15,I4,I4,I4)
1001  FORMAT(1X,8(I4),'  N M L NPROCX NPROCY NPROCZ',
     1                ' PVMROUTE ENCODING')
      RETURN
      END
